@@ -1,20 +1,20 @@
-#!/usr/bin/perl -w
+#!/usr/bin/perl -Tw
 
 # Copyright (c) 2008, Dylan Doxey.  All rights reserved. This program
 # is free software; you may use it under the same terms as Perl
 # itself.
 #
-# Revised copy of Lingua::Translate::Babe lfish by Sam Vilain
+# Revised copy of Lingua::Translate::Babelfish by Sam Vilain
 # <enki@snowcra.sh>
 #
 package Lingua::Translate::Google;
 
-our $VERSION = '0.03';
+our $VERSION = '0.04';
 
 use strict;
 use Carp;
 use LWP::UserAgent;
-use HTTP::Request::Common qw(GET);
+use HTTP::Request::Common qw( GET POST );
 use Unicode::MapUTF8 qw(to_utf8);
 use I18N::LangTags qw(is_language_tag);
 
@@ -35,9 +35,15 @@ sub new {
     is_language_tag( $self->{src} = delete $options{src} )
       or croak "$self->{src} is not a valid RFC3066 language tag";
 
+    $self->available( $self->{src} )
+        or croak "$self->{dest} is not available on Google translate"; 
+
     is_language_tag( $self->{dest} = delete $options{dest} )
       or croak "$self->{dest} is not a valid RFC3066 language tag";
 
+    $self->available( $self->{dest} )
+        or croak "$self->{dest} is not available on Google translate"; 
+
     $self->config(%options);
 
     return $self;
@@ -51,10 +57,6 @@ sub translate {
     my $text = shift;
 
     # TODO Determine if chunkification is necessary for Google.
-    # chunkify the text.  Knowing how to do this properly is really
-    # the job of an expert system, so I'll keep it simple and break on
-    # English sentence terminators (which might be completely the
-    # wrong thing to do for other languages.  Oh well)
     my @chunks = (
         $text =~ m/\G\s*   # strip excess white space
 			     (
@@ -65,10 +67,10 @@ sub translate {
 			         # string
 			         (?:[\.;:!\?]|$)
 			     )
-			     /xsg
+        /xsg
     );
     die 'Could not break up given text into chunks'
-      if ( pos($text) and pos($text) < length($text) );
+        if ( pos($text) and pos($text) < length($text) );
 
     # the translated text
     my ( @translated, $error );
@@ -76,55 +78,57 @@ sub translate {
     CHUNK:
     for my $chunk (@chunks) {
 
-        my $service_uri = $self->{google_uri};
+        my $req;
 
-        # for http://ajax.googleapis.com/ajax/services/language/translate?
-        my %params = (
-            v        => '1.0',
-            key      => $self->{api_key},
-            q        => $chunk,
-            langpair => $self->{src} . '%7C' . $self->{dest},
-        );
+        # Primary service: google_uri
+        if ( $self->{api_key} ) {
 
-        # for http://www.google.com/uds/Gtranslate?
-        ## my %params = (
-        ##     v        => '1.0',
-        ##     key      => $self->{api_key},
-        ##     q        => $chunk,
-        ##     langpair => $self->{src} . '%7C' . $self->{dest},
-        ##     context  => 22
-        ##     callback => 'google.language.callbacks.id101',
-        ## );
-
-        # for http://translate.google.com/translate_a/t?
-        if ( !$self->{api_key} ) {
-
-            %params = (
-                text   => $chunk,
-                sl     => $self->{src},
-                tl     => $self->{dest},
-                client => 't',
+            my %params = (
+                v        => '1.0',
+                key      => $self->{api_key},
+                q        => $chunk,
+                langpair => $self->{src} . '%7C' . $self->{dest},
             );
-            $service_uri = $self->{google_fallback_uri};
+            my $query = join '&', map { "$_=$params{$_}" } keys %params;
 
-            # Note, for the fallback service might return windows-1255
-            # encoding, unless User-Agent suggests a Linux client.
+            $req = GET "$self->{google_uri}$query";
+
+            if ( !$self->{referer} ) {
+
+                # Warn because Google surely frowns upon doing this.
+                warn "no valid referer provided, using $self->{google_uri}";
+                $self->{referer} = $self->{google_uri};
+            }
         }
 
-        my $query = join '&', map { "$_=$params{$_}" } keys %params;
+        # Fallback service at: google_fallback_uri
+        else {
 
-        # make a new request object
-        my $req = GET "$service_uri$query";
+            my %params = (
+                client => 't',
+                sl     => $self->{src}, 
+                tl     => $self->{dest},
+                text   => $chunk,
+            );
+            my $query = join '&', map { "$_=$params{$_}" } keys %params;
+
+            $req = GET "$self->{google_fallback_uri}$query";
+
+            if ( $self->{referer} && $self->{referer} !~ m/translate[.]google[.]com/msx ) {
+
+                # Warn because Google surely frowns upon doing this.
+                warn "overwriting referer $self->{referer} to $self->{google_fallback_uri}\n";
+                $self->{referer} = $self->{google_fallback_uri};
+            }
+        }
 
         $req->header( 'Referer',        $self->{referer} );
         $req->header( 'Accept-Charset', 'UTF-8' );
 
-      RETRY:
-
         # try several times to reach google
+        RETRY:
         for my $attempt ( 1 .. $self->{retries} + 1 ) {
 
-            # go go gadget LWP::UserAgent
             my $res = $self->agent->request($req);
 
             if ( $res->is_success ) {
@@ -135,11 +139,9 @@ sub translate {
                 );
 
                 # google errors
-                # FIXME Get the correct message text
+                # TODO Get the correct message text
                 next RETRY
-                  if $output =~ m/ We \s+ were \s+ unable \s+
-                    to \s+ process \s+ your \s+ request \s+
-                    within \s+ the \s+ available \s+ time/msx;
+                    if $output =~ m/ Out \s+ of \s+ commission /msx;
 
                 # trim
                 $output =~ s/(?: \A \s* | \s* \z )//msxg;
@@ -156,11 +158,9 @@ sub translate {
 
         # give up
         die "Request timed out more than $self->{retries} times ($error)";
-
     }
 
     return join ' ', @translated;
-
 }
 
 # Extracts the translated text from the given Google response text.
@@ -202,39 +202,54 @@ sub _extract_text {
 
 # Returns the available language translation pairs.
 sub available {
-
     my $self = shift;
+    my ($lang_inquiry) = @_;
+
     UNIVERSAL::isa( $self, __PACKAGE__ )
       or croak __PACKAGE__ . '::available() called as function';
 
+    my $uri = $self->{google_translate_uri};
+
     # return a cached result
-    if ( my $ok_langs = $valid_langs{ $self->{google_uri} } ) {
-        return keys %$ok_langs;
+    if ( my $ok_langs = $valid_langs{ $uri } ) {
+
+        return $self->_has_language( $lang_inquiry )
+            if $lang_inquiry;
+
+        return keys %{ $ok_langs };
     }
 
     # create a new request
-    my $req = GET $self->{google_uri};
+    my $req = GET $uri;
+    $req->header( 'Referer', $uri );
+    $req->header( 'Accept-Charset', 'UTF-8' );
 
-    # go get it
     my $res = $self->agent->request($req);
-    die 'Google fetch failed; ' . $res->status_line
-      unless $res->is_success;
+
+    die 'Google fetch failed; ' . $res->status_line()
+        unless $res->is_success();
 
     # extract out the languages
-    my $page = $res->content;
+    my $page = $res->content();
 
     my @source_langs;
     if ( $page =~ m{<select \s+ name=sl [^>]+ > ( .*? ) </select>}msx ) {
+
         my $options = $1;
-        while ( $options =~ m{<option \s+ value="([^"]+)">}msx ) {
+
+        while ( $options =~ m{<option \s+ value="([^"]+)">}msxg ) {
+
             my $lang = $1;
             push @source_langs, $lang;
         }
     }
     my @list;
     if ( $page =~ m{<select \s+ name=tl [^>]+ > ( .*? ) </select>}msx ) {
+
         my $options = $1;
-        while ( $options =~ m{<option \s+ value="([^"]+)">}msx ) {
+
+        while ( $options =~ m{<option \s+ value="([^"]+)">}msxg ) {
+
             my $dest = $1;
 
             # Presumably any source language is paired with any to language.
@@ -251,10 +266,32 @@ sub available {
         }
     }
 
-    # save the result
-    $valid_langs{ $self->{google_uri} } = map { $_ => 1 } @list;
+    if ( @list > 0 ) {
 
-    return @list;
+        # save the result
+        %{ $valid_langs{ $uri } } = map { $_ => 1 } @list;
+
+        return $self->_has_language( $lang_inquiry )
+            if $lang_inquiry;
+
+        return @list;
+    }
+    warn "unable to parse valid language tokens from $uri";
+    return;
+}
+
+sub _has_language {
+    my $self = shift;
+    my ($lang_inquiry) = @_;
+
+    my $lang_regex = qr{ (?: \A $lang_inquiry _ | _ $lang_inquiry \z ) }msx;
+
+    for my $lang_pair (keys %{ $valid_langs{ $self->{google_translate_uri} } }) {
+
+        return 1
+            if $lang_pair =~ $lang_regex;
+    }
+    return 0; 
 }
 
 # Returns the LWP::UserAgent object.
@@ -301,6 +338,10 @@ sub config {
 
             $self->{google_fallback_uri} = $value;
         }
+        elsif ( $option eq 'google_translate_uri' ) {
+
+            $self->{google_translate_uri} = $value;
+        }
         elsif ( $option eq 'api_key' ) {
 
             $self->{api_key} = $value;
@@ -323,6 +364,10 @@ sub config {
 
             $self->{chunk_size} = $value;
         }
+        elsif ( $option eq 'use_cache' ) {
+
+            $self->{use_cache} = $value;
+        }
         elsif ( $option eq 'retries' ) {
 
             $self->{retries} = $value;
@@ -343,6 +388,7 @@ use Pod::Constants
             my ($varname) = m/(\w+)/;
 
             my ($default) = m/The default value is\s+"(.*?)"/s;
+
             config( $varname => $default );
         }
     );
@@ -371,14 +417,16 @@ Lingua::Translate::Google - Translation back-end for Google's beta translation s
 
 =head1 SYNOPSIS
 
+ # 
+ # Translate a string.
+ # 
  use Lingua::Translate;
 
  Lingua::Translate::config
      (
-       backend => 'Google',
-       api_key => 'YoUrApIkEy',
-       referer => 'http://your.domain.tld/yourdir/',
-       ua      => LWP::UserAgent->new(),
+         back_end => 'Google',
+         api_key  => 'YoUrApIkEy',
+         referer  => 'http://your.domain.tld/yourdir/',
      );
 
  my $xl8r = Lingua::Translate->new(src => 'de', dest => 'en');
@@ -387,33 +435,82 @@ Lingua::Translate::Google - Translation back-end for Google's beta translation s
  print $xl8r->translate('Mein Luftkissenfahrzeug ist voller Aale');
 
 
+ # 
+ # Perhaps you'd like to build a collection of translations.
+ # 
+ use Lingua::Translate;
+ use Data::Dumper;
+
+ my $src = 'en';
+
+ my %source_lang = (
+     hovercraft_eels    => 'My hovercraft is full of eels.',
+     cigarettes_matches => 'I would like some cigarettes and a box of matches.',
+     hello_world        => 'Hello world.'
+ );
+
+ my %result_lang;
+
+ Lingua::Translate::config
+     (
+         back_end => 'Google',
+         api_key  => 'YoUrApIkEy',
+         referer  => 'http://your.domain.tld/yourdir/',
+     );
+
+ DEST:
+ for my $dest (qw( iw pt ro fr de hi es ja zh-CN )) {
+
+     my $xl8r = Lingua::Translate->new(
+         src      => $src,
+         dest     => $dest,
+     ) or die "No translation server available for $src -> $dest";
+
+     TOKEN:
+     for my $token ( keys %source_lang ) {
+
+         my $source_text = $source_lang{$token};
+
+         $result_lang{$dest}->{$token} = $xl8r->translate($source_text);
+     }
+ }
+
+ print Dumper( \%result_lang ) . "\n";
+
+
 =head1 DESCRIPTION
 
-Lingua::Translate::Google is a translation back-end for Lingua::Translate that contacts Google (http://ajax.googleapis.com/ajax/services/language/translate/) to do the real work.
+Lingua::Translate::Google is a translation back-end for Lingua::Translate that contacts Google translation service to do the real work.
+The Google translation API is currently at L<http://ajax.googleapis.com/ajax/services/language/translate/.>
 
-It is normally invoked by Lingua::Translate; there should be no need to call it directly.  If you do call it directly, you will lose the ability to easily switch your programs over to alternate back-ends that are later produced.
+Lingua::Translate::Google is normally invoked by Lingua::Translate; there should be no need to call it directly.  If you do call it directly, you will lose the ability to easily switch your programs over to alternate back-ends that are later produced.
 
-If you omit the API key config option, then this module uses the fallback service at translate.google.com/ which works fine without it.
+If you omit the API key config option, then this module uses the fallback service at translate.google.com/ which works fine without it, on the condition that the referer is the URL of the translate service. 
 
-To use the Google APIs, Google asks that you obtain an API key, and that you always include a valid and accurate referer URL. If you supply an API key, then this module uses the AJAX API which Google provides specifically for third party application development.
-
-One way or another, by using Google services (either directly or via this module) you are agreeing by their terms of service.
 
 =over
 
 =item Please read:
 
-http://www.google.com/accounts/TOS
+By using Google services (either directly or via this module) you are agreeing by their terms of service.
 
-=item To obtain your own API key:
+L<http://www.google.com/accounts/TOS>
+
+
+=item To obtain your API key:
+
+To use the Google APIs, Google asks that you obtain an API key, and that you always include a valid and accurate referer URL. If you supply an API key, then this module uses the AJAX API which Google provides specifically for third party application development.
+
+
+L<http://code.google.com/apis/ajaxfeeds/signup.html>
 
-http://code.google.com/apis/ajaxfeeds/signup.html
 
 =back
 
 
 =head1 CONSTRUCTOR
 
+
 =head2 new(src => $lang, dest => lang)
 
 Creates a new translation handle.
@@ -444,9 +541,15 @@ The following methods may be called on Lingua::Translate::Google objects.
 Returns a list of available language pairs, in the form of 'XX_YY', where XX is the source language and YY is the destination.
 If you want the english name of a language tag, call I18N::LangTags::List::name() on it.  See L<I18N::LangTags::List>.
 
-This method contacts Google (http://translate.google.com/translate_a/t?) to calculate available language pairs.
-The data about available language pairs is cached in memory.
+This method contacts Google (at the configured google_fallback_uri) and parses from the HTML the available language pairs. The list of language pairs is cached for subsequent calls.
 
+As of Lingua::Translate version 0.09, calls to this method don't propogate from the Lingua:Translate namespace.
+Rather, this method is only available in the Lingua::Translate::Google namespace.
+
+You may also use this method to see if a given language tag is available.
+
+ die "doesn't have 'he'"
+     if !$xl8tr->available('he');
 
 =head2 translate($text) : $translated
 
@@ -471,6 +574,20 @@ This function sets defaults for use when constructing objects. Options include:
 
 =over
 
+=item api_key
+
+This key is issued to you by Google and it gives you access to the AJAX API which is the translation service intended for application development.
+
+See: L<http://code.google.com/apis/ajaxfeeds/signup.html>
+
+=item referer
+
+The value for the referer header in HTTP requests sent to the Google translation service.
+
+Google requests that you provide a valid referer string. You will probably use the one you specified when you got your API key. If you're not specifying an API key, then the fall-back translator will reject (with a "403 Forbidden" response) any request which does not have the translator URL as its referer. 
+
+The translate function will make sure you have a referer value that works, and will warn if it's overwriting the value you specified. 
+
 =item google_uri
 
 The uri to use when contacting Google.
@@ -501,7 +618,7 @@ Another (yet currently unsported) possibility is:
 
 =item google_fallback_uri
 
-The URI used when contacting Google, and no api_key is provided.
+The URI used when contacting Google and no api_key is provided. This is the AJAX service used by the public translate site. 
 
 The default value is
 
@@ -514,6 +631,14 @@ The default value is
 
 Note, Google states clearly that they want you to obtain and use an API key, and also include a valid and accurate referer URL.
 
+=item google_translate_uri
+
+This is the URL of the Google page where the available languages are parsed from. The L<item available()_:_@list> method uses this URI to get some HTML which is presumed to have a select list for the source and destination languages.
+
+The default value is
+
+"http://translate.google.com/translate_t#"
+
 =item agent
 
 The User-Agent string to use when contacting Google.
@@ -530,8 +655,12 @@ The number of times to retry contacting Google if the first attempt fails. The d
 
 =back
 
+=head1 DIAGNOSTICS
 
-=head1 BUGS/TODO
+Expect to see a warning if you're using the api_key and omitting the referer. Also, expect to see a warning if you're omitting the api_key and specifying a custom referer. This is because the Google translation service cares about your referer values. This module will warn whenever it's changing what you specified to make the translation work.
+
+
+=head1 TODO
 
 The chunk_size attribute is a hold-over from the Babelfish algorithm. It is TBD as to what chunk size ought to be set for Google.
 
@@ -543,13 +672,18 @@ There might be a better way to get the available language pairs.
 L<Lingua::Translate>, L<Lingua::Translate::Babelfish>, L<LWP::UserAgent>, L<Unicode::MapUTF8>
 
 
+=head1 LICENSE
+
+This is free software, and can be used/modified under the same terms as Perl itself.
+
 =head1 ACKNOWLEDGEMENTS
 
-This is a rewritten copy of Lingua::Translate::Babelfish by Sam Vilain <enki@snowcra.sh>.
+Sam Vilain (L<http://search.cpan.org/~samv/>) wrote Lingua::Translate::Babelfish which served as the basis for this module.
 
+Jerrad Pierce (L<http://search.cpan.org/~jpierce/>) for bug reporting.
 
 =head1 AUTHOR
 
-Dylan Doxey, <dylan.doxey@gmail.com>
+Dylan Doxey, <dylan@cpan.org>
 
 =cut
